# Description

Shows a friendly page to the users when an exception or an error are thrown,
instead of the plain page provided by Drupal core. The module is able to attach
a Universally Unique Identifier (UUID) to each error/exception, so that the
user is able to refer the error to the site's contact support team or help desk.
The UUID is logged also in the Drupal log.

# Technical aspects

## Drupal services, configuration, theming

Uncaught exceptions indicate that something really wrong has happened on a
site. There is a great chance that some, or all, of Drupal services are not
available. For this reason, in the module we are avoiding use or refer Drupal
services as much as possible. For the same reason, the module cannot be
configured using the configuration management offered by Drupal. Instead, the
module will use the settings stored in the `settings.php` file to configure its
behavior.

This is true also for theming. The Drupal/Twig rendering engine cannot be used,
as it would involve usage of Drupal services as well. However, the outputted
page or message can still be customized because they are simple static HTML
pages with some raw replacement tokens.

## How does it work?

### Uncaught exceptions

The module swaps the core `ExceptionLoggingSubscriber`, which is responsible
for logging the uncaught exceptions to the Drupal logger, with its own
`ErrorPageExceptionLoggingSubscriber`. Actually the module class replaces the
`::onError()` method and logs also the UUID if it has been configured so. Then,
the module adds a new subscriber, `ErrorPageFinalExceptionSubscriber`, which
acts just before the core `FinalExceptionSubscriber` by showing the output in
its own way and stopping the propagation of the `KernelEvents::EXCEPTION` event.

### Fatal and user errors

The module uses its own error/exception handlers, instead of
`_drupal_error_handler()` and `_drupal_exception_handler()`. Because those
procedural functions are hard to be extended or overwritten, the core code is
95% copied in module's handlers. These are slightly changed to allow using a
custom-outputted HTML rendered markup.

# Getting started

## Install

Install the module as any other module. There are no other dependencies.

## Configure

As explained earlier, the entire configuration is done via `settings.php`:

### ErrorPageErrorHandler autoload

As the container might not be available yet, when an uncaught exception or a
fatal error occurs, the autoloading might not work for extensions such as
modules. Thus, we cannot register our error/exception handlers, which live in
`ErrorPageErrorHandler`. The system must be instructed explicitly from where to
load the class. If you are using composer this is done automatically because in
the module's `composer.json` we've added the class as a `classmap` entry of
`autoload`. If, for some reason, you're not using Composer, then you'll need to
explicitly require the file. In this case, in `settings.php` just add this line:

```php
// Note that the path to error handler might be different on some installations.
require_once 'modules/contrib/error_page/src/ErrorPageErrorHandler.php';
```

### Set the custom error handlers

Next two lines should be added in `settings.php`:

```php
set_error_handler(['Drupal\error_page\ErrorPageErrorHandler', 'handleError']);
set_exception_handler([
  'Drupal\error_page\ErrorPageErrorHandler',
  'handleException',
]);
```

### Configuration

In `settings.php`:

```php
// If to create a UUID for each exception/error. Defaults to TRUE.
$settings['error_page']['uuid'] = TRUE;
// Point to the path where the customizable HTML markup files are placed. It's
// recommended that the custom template location is placed outside the web tree
// or is protected from the public access with a file, such as markup/.htaccess.
// Defaults to the templates provided by module, in the `markup/` subdirectory.
$settings['error_page']['template_dir'] = DRUPAL_ROOT . '/../path/to/templates';
```

### All together

Your `settings.php` section might look like:

```php
// Only if you don't use composer.
require_once 'modules/contrib/error_page/src/ErrorPageErrorHandler.php';
set_error_handler(['Drupal\error_page\ErrorPageErrorHandler', 'handleError']);
set_exception_handler([
  'Drupal\error_page\ErrorPageErrorHandler',
  'handleException',
]);
// Log the UUID in the Drupal logs.
$settings['error_page']['uuid'] = TRUE;
// Your templates are located in path/to/templates, one level above the webroot.
$settings['error_page']['template_dir'] = DRUPAL_ROOT . '/../path/to/templates';
```

# Customizing the output page/message

By default, the HTML markup page and message, displayed in case of error or
exception, are basic, but they can be customized. Probably the word "themed"
would be too much. In order to customize the output, the files
`markup/error_page.html` and `markup/error_message.html` should be copied and
edited in the location specified in `$settings['error_page']['template_dir']`.
Three variables can be used:

- `{{ uuid }}`: The error/exception UUID, if any.
- `{{ base_path }}`: The Drupal base path, as is returned by `base_path()`. This
  helps to build paths to images or other assets.
- `{{ error_report }}`: The standard error and the error/exception backtrace.
  Note that the content of this variable is empty if the system error logging
  verbosity doesn't allow showing verbose error details such as the backtrace.

Remember to protect the template location from public access or, even better,
place them outside the web tree.

## Testing

The package contains a module, `error_page_test`, that simulates different kind
of exceptions, errors or notices. As this is a testing module, before enabling,
make sure Drupal allows discovery of testing module by ensuring this line in
`settings.local.php` or `settings.php`:

```php
$settings['extension_discovery_scan_tests'] = TRUE;
```

_Warning_: You should never enable `extension_discovery_scan_tests` setting in
production. Discovering testing modules should be enabled only on development
environments.

After installing `error_page_test`, you can test the following pages:

* `/error_page_test/exception`: simulates an exception.
* `/error_page_test/fatal_error`: simulates a fatal error.
* `/error_page_test/user_error`: simulates a user error.
* `/error_page_test/php_notice`: simulates a PHP notice.

By default, exceptions and fatal errors are logged also in the PHP log. For
testing purposes, you can pipe the error log entries in a custom file:

```php
// Check https://www.php.net/manual/en/function.error-log.php for details.
// A value of 3 sends the error log entries in a file.
$settings['error_page']['log']['method'] = 3;
// Defines the destination file.
$settings['error_page']['log']['destination'] = 'temporary://error_log.txt';
```

## Contributing

[DDEV](https://ddev.com), a Docker-based PHP development tool for a streamlined
and unified development process, is the recommended tool for contributing to the
module. The [DDEV Drupal Contrib](https://github.com/ddev/ddev-drupal-contrib)
addon makes it easy to develop a Drupal module by offering the tools to set up
and test the module.

### Install DDEV

* Install a Docker provider by following DDEV [Docker Installation](https://ddev.readthedocs.io/en/stable/users/install/docker-installation/)
  instructions for your Operating System.
* [Install DDEV](https://ddev.readthedocs.io/en/stable/users/install/ddev-installation/),
  use the documentation that best fits your OS.
* DDEV is used mostly via CLI commands. [Configure shell completion &
  autocomplete](https://ddev.readthedocs.io/en/stable/users/install/shell-completion/)
  according to your environment.
* Configure your IDE to take advantage of the DDEV features. This is a critical
  step to be able to test and debug your module. Remember, the website runs inside Docker, so pay attention to these configurations:
  - [PhpStorm Setup](https://ddev.readthedocs.io/en/stable/users/install/phpstorm/)
  - [Configure](https://ddev.readthedocs.io/en/stable/users/debugging-profiling/step-debugging/)
    PhpStorm and VS Code for step debugging.
  - Profiling with [xhprof](https://ddev.readthedocs.io/en/stable/users/debugging-profiling/xhprof-profiling/),
    [Xdebug](https://ddev.readthedocs.io/en/stable/users/debugging-profiling/xdebug-profiling/)
    and [Blackfire](https://ddev.readthedocs.io/en/stable/users/debugging-profiling/blackfire-profiling/).

### Checkout the module

Normally, you check out the code form an [issue fork](https://www.drupal.org/docs/develop/git/using-gitlab-to-contribute-to-drupal/creating-issue-forks):

```shell
git clone git@git.drupal.org:issue/error_page-[issue number].git
cd error_page-[issue number]
```

### Start DDEV

Inside the cloned project run:

```shell
ddev start
```

This command will fire up the Docker containers and add all configurations.

### Install dependencies

```shell
ddev poser
```

This will install the PHP dependencies. Note that this is a replacement for
Composer _install_ command that knows how to bundle together Drupal core and the
module. Read more about this command at
https://github.com/ddev/ddev-drupal-contrib?tab=readme-ov-file#commands

```shell
ddev symlink-project
```

This symlinks the module inside `web/modules/custom`. Read more about this
command at https://github.com/ddev/ddev-drupal-contrib?tab=readme-ov-file#commands.
Note that as soon as `vendor/autoload.php` has been generated, this command runs
automatically on every `ddev start`.

This command should also be run when adding new directories or files to the root
of the module.

```shell
ddev exec "cd web/core && yarn install"
```

Install Node dependencies. This is needed for the `ddev eslint` command.

### Install Drupal

```shell
ddev install
```

This will install Drupal and will enable the module.

### Changing the Drupal core version

* Create a file `.ddev/config.local.yaml`
* Set the desired Drupal core version. E.g.,
  ```yaml
  web_environment:
    - DRUPAL_CORE=^10.3
  ```
* Run `ddev restart`

### Run tests

* `ddev phpunit`: run PHPUnit tests
* `ddev phpcs`: run PHP coding standards checks
* `ddev phpcbf`: fix coding standards findings
* `ddev phpstan`: run PHP static analysis
* `ddev eslint`: Run ESLint on Javascript and YAML files.
