<?php

namespace Drupal\error_page;

use Drupal\Component\Uuid\Php;
use Drupal\Core\Site\Settings;
use Drupal\Core\Utility\Error;

/**
 * Custom fatal error handler.
 */
class ErrorPageErrorHandler {

  /**
   * Handles errors.
   *
   * Code copied and adapted from _drupal_error_handler_real().
   *
   * @param int $error_level
   *   Error level.
   * @param string $message
   *   Error message.
   * @param string $filename
   *   Error file.
   * @param int $line
   *   Error line in file.
   *
   * @see _drupal_error_handler_real()
   */
  public static function handleError($error_level, $message, $filename, $line): void {
    require_once DRUPAL_ROOT . '/core/includes/errors.inc';
    require_once __DIR__ . '/../errors.inc';
    // @codingStandardsIgnoreLine
    /** @phpstan-ignore function.notFound */
    _error_page__drupal_error_handler_real($error_level, $message, $filename, $line);
  }

  /**
   * Handles exceptions.
   *
   * Code copied and adapted from _drupal_exception_handler().
   *
   * @param \Throwable $exception
   *   The throwable.
   *
   * @see _drupal_exception_handler()
   */
  public static function handleException(\Throwable $exception) {
    require_once DRUPAL_ROOT . '/core/includes/errors.inc';

    try {
      // Log the message to the watchdog and return an error page to the user.
      static::logError(Error::decodeException($exception), TRUE);
    }
    // Catch \Throwable, which covers both Error and Exception throwables.
    catch (\Throwable $error) {
      _drupal_exception_handler_additional($exception, $error);
    }
  }

  /**
   * Logs a PHP error or exception and displays an error page in fatal cases.
   *
   * This method is a slight changed version of _drupal_log_error().
   *
   * @param array $error
   *   An array with the following keys: %type, @message, %function, %file,
   *   %line, @backtrace_string, severity_level, backtrace, and exception. All
   *   the parameters are plain-text, with the exception of @message, which
   *   needs to be an HTML string, backtrace, which is a standard PHP backtrace,
   *   and exception, which is the exception object (or NULL if the error is not
   *   an exception).
   * @param bool $fatal
   *   TRUE for:
   *   - An exception is thrown and not caught by something else.
   *   - A recoverable fatal error, which is a fatal error.
   *   Non-recoverable fatal errors cannot be logged by Drupal.
   *
   * @see _drupal_log_error()
   */
  protected static function logError(array $error, $fatal = FALSE): void {
    require_once DRUPAL_ROOT . '/core/includes/errors.inc';
    require_once __DIR__ . '/../errors.inc';
    $settings = Settings::get('error_page');
    $uuid_enabled = $settings['uuid'] ?? TRUE;
    $uuid = $uuid_enabled ? (new Php())->generate() : NULL;
    $error['@uuid'] = $uuid;
    // @codingStandardsIgnoreLine
    /** @phpstan-ignore function.notFound */
    _error_page__drupal_log_error($error, $fatal);
  }

}
